-- Modellierung von while-Programmen durch
-- Haskell-Datenstrukturen (analog zur
-- Fallstudie in Kapitel 6 der Vorlesung):

------------------------------------------------------------
-- Konstruktion der Syntax der while-Sprache in Datentypen:

-- Es werden die vier mglichen Anweisungsarten - leere und
-- bedingte Anweisung, Schleife und Zuweisung - im Datentyp
-- Anweisung definiert und ferner die Typen Bedingung und
-- Ausdruck sowie die Operatoren eingefhrt.

data Operator  = Plus | Minus | Mal | Div | Mod

haskellOp :: Operator -> (Int->Int->Int)
haskellOp Plus  = (+)
haskellOp Minus = (-)
haskellOp Mal   = (*)
haskellOp Div   = div
haskellOp Mod   = mod

data Vergleicher = Kl | Gl | Gr | Kg | Ug | Gg

haskellRop :: Vergleicher -> (Int->Int->Bool)
haskellRop Kl = (<)      -- Kleiner als
haskellRop Gl = (==)     -- Gleich
haskellRop Gr = (>)      -- Grer als
haskellRop Kg = (<=)     -- Kleiner oder gleich
haskellRop Ug = (/=)     -- Ungleich
haskellRop Gg = (>=)     -- Grer oder gleich

-- Die Funktionen HaskellOp und HaskellRop, sowie die Daten-
-- typen Operator und Vergleicher dienen dazu, um aus den
-- Verknpfungstoken im Zustandsbaum die arithmetischen bzw.
-- die Vergleichsfunktionen lesen und verwerten zu knnen.

-- Aufruf:   haskellOp Minus
-- Ergebnis: <<function>>
-- Aufruf:   (haskellOp Plus) 1 2
-- Ergebnis: 3
-- Aufruf:   haskellRop Ug 1 2
-- Ergebnis: True

data Bedingung =
   Wahr Bool | Vergleich Ausdruck Vergleicher Ausdruck

data Ausdruck =
   Zahl Int | Name String | Verkn Ausdruck Operator Ausdruck

data Anweisung =
   Fall Bedingung [Anweisung] [Anweisung] |
   Schleife Bedingung [Anweisung] |
   Skip |
   Zuweisung String Ausdruck

------------------------------------------------------------
-- Modellierung der Abstrakten Maschine in Haskell:

-- Zunchst wird die Syntax der Zustnde, bestehend aus der
-- aktuellen Umgebung und den gleich aufgebauten Werte- und
-- Programmkellern, modelliert; dann Beispielkonfigurationen
-- erstellt und zuletzt delta als die eigentliche Zustands-
-- bergangsfunktion eingefhrt.

data Wert = Undef | Wert Int

istUndef :: Wert -> Bool
istUndef Undef = True
istUndef _     = False

-- istUndef ist eine Prffunktion fr eine Variable in der
-- Zustandsumgebung, welche Wahr als Boolean ausgibt, wenn
-- die Variable keinen Wert hat, also noch undefiniert ist.
-- Die Funktion basiert auf dem algebraischen Datentyp Wert.

-- Aufruf:   istUndef Undef
-- Ergebnis: True
-- Aufruf:   istUndef (Wert 3)
-- Ergebnis: False

data Variable = Variable String Wert

varD :: Variable
varD = Variable "d" (Wert 0)
-- Die Variable d wird deklariert.

type Umgebung = [Variable]
-- Der Datentyp Umgebung, als Liste von Variablen, ist
-- die erste Komponente eines Zustands in der abstrakten
-- Maschine des While-Interpreters.

wertSuche :: String -> Umgebung -> Wert
wertSuche s ((Variable v wert):u)
  | eqString s v = wert
  | otherwise    = wertSuche s u
wertSuche _ []   = Undef

-- Mit wertSuche kann aus einer Umgebung im aktuellen
-- Zustand der Wert einer Variable ausgegeben werden.

-- Aufruf:   wertSuche "a" [(Variable "b" (Wert 0)),
--                          (Variable "c" (Undef)),
--                          (Variable "a" (Wert 5))]
-- Ergebnis: (Wert 5)
-- Aufruf:   wertSuche "a" [(Variable "b" (Wert 0))]
-- Ergebnis: Undef

wertEintrag :: String -> Int -> Umgebung -> Umgebung
wertEintrag s n ((Variable v wert):u)
  | eqString s v   = (Variable s (Wert n)): u
  | otherwise      = (Variable v wert):(wertEintrag s n u)
wertEintrag s _ [] = error ("Variable "++s++" nicht in Umgebung")

-- wertEintrag wird fr die Regel 10 der Zustandsbergangs-
-- funktion verwendet, welche den Integerwert einer Variable
-- in der aktuellen Umgebung verndert, bzw. im Falle Undef
-- erstmalig belegt. Es kann nur auf bereits in der aktuellen
-- Umgebung vorhandene Variablen zugegriffen werden.

-- Aufruf:   wertEintrag "c" 1 [(Variable "b" (Wert 0)),
                                (Variable "c" (Undef)),
                                (Variable "a" (Wert 5))]
-- Ergebnis:                   [(Variable "b" (Wert 0)),
                                (Variable "c" (Wert 1)),
                                (Variable "a" (Wert 5))]
-- Aufruf:   wertEintrag "a" 1 [(Variable "b" (Wert 0))]
-- Ergebnis: exception: Variable a nicht in Umgebung.

um :: Umgebung
um = [(Variable "d" Undef),
      (Variable "x" (Wert 17)),
      (Variable "y" (Wert 4))]
-- Die Umgebung um wird deklariert.

-- Inhalt des Wertekellers und des Programmkellers:

data Teil =
   Anw Anweisung |
   AnwF [Anweisung] |
   Bed Bedingung |
   Ausdr Ausdruck |
   Bez String |
   Zwert Int |
   Wwert Bool |
   Op Operator |
   Rop Vergleicher |
   If |
   Do |
   Zuw
-- Der Datentyp Teil definiert die Kellerkomponenten
-- der Zustnde. Dies gibt auch gleichzeitig die Syntax
-- vor, in welche die while-Programme umgeformt (per Hand
-- oder Compiler-Compiler) werden mssen, damit der While-
-- Interpreter sie auslesen und verwerten kann.

type Konfiguration =
   (Umgebung, [Teil], [Teil])
-- Der komplette Zustand mit seinen drei Komponenten.

data Zustand = Ende Umgebung | Akt Konfiguration
-- Der Datentyp Zustand, um vom Anfangszustand sowie den
-- den Zwischenzustnden whrend der Auswertung und von
-- der Endspeicherbelegung, fr die die Kellerinhalte nicht
-- mehr von Bedeutung sind, unterscheiden zu knnen.

------------------------------------------------------------
-- In den folgenden fuenf Beispielkonfigurationen
-- wurden die Haskell-Darstellungen der while-
-- Programme nicht von Hand, sondern mit Hilfe
-- des Compiler-Compilers SIC erzeugt (wie in der
-- Vorlesung vorgefuehrt).

-- Nach jeder Beispielkonfiguration wird das While-Programm
-- in der While-Syntax als Kommentar angefgt.

-- !!!!!!!! Wie ist das mit den Semikolon in der while sprache? bitte die comments ggf verbessern


drStart1 :: Konfiguration
drStart1 =
  ([(Variable "d" Undef),
    (Variable "x" (Wert 17)),
    (Variable "y" (Wert 4))],
    [],
    [AnwF [(Zuweisung "d" (Zahl 0))]]
  )

z1 :: Zustand
z1 = Akt drStart1

-- d := 0;

-- Einfache Zuweisung

drStart2 :: Konfiguration
drStart2 =
  ([(Variable "d" (Wert 1)),
    (Variable "x" (Wert 17)),
    (Variable "y" (Wert 4)),
    (Variable "z" Undef)],
    [],
    [AnwF [
      (Zuweisung "d" (Verkn (Name "d") Plus (Zahl 1))),
      (Zuweisung "x" (Verkn (Name "x") Minus (Name "y")))]
    ]
  )

z2 :: Zustand
z2 = Akt drStart2

-- d := (d+1);
-- x := (x-y);

-- Einfache arithmetische Ausdrcke

drStart3 :: Konfiguration
drStart3 =
  ([(Variable "d" Undef),
    (Variable "x" (Wert 17)),
    (Variable "y" (Wert 4))],
    [],
    [AnwF [(Zuweisung "d" (Zahl 0)),
           (Schleife (Vergleich (Name "x") Gg (Name "y"))
           [(Zuweisung "d" (Verkn (Name "d") Plus (Zahl 1))),
            (Zuweisung "x" (Verkn (Name "x") Minus (Name "y")))
           ])
          ]]
  )

z3 :: Zustand
z3 = Akt drStart3

-- d := 0;
-- while (x>=y) do
--   d := (d+1);
--   x := (x-y);
-- od

-- Division mit Rest

drStart4 :: Konfiguration
drStart4 =
  ([(Variable "f" Undef),
    (Variable "n" (Wert 6))],
    [],
    [AnwF [(Zuweisung "f" (Zahl 1)),
           (Schleife (Vergleich (Name "n") Gr (Zahl 1))
            [(Zuweisung "f" (Verkn (Name "f") Mal (Name "n"))),
             (Zuweisung "n" (Verkn (Name "n") Minus (Zahl 1)))
            ])
          ]]
  )

z4 :: Zustand
z4 = Akt drStart4

-- f := 1;
-- while (n>1) do
--   f:=(f*n);
--   n:=(n-1);
-- od

-- Fakulttsfunktion

drStart5 :: Konfiguration
drStart5 =
  ([(Variable "f" Undef),
    (Variable "n" (Wert 0))],
    [],
    [AnwF [(Zuweisung "f" (Zahl 1)),
           (Schleife (Vergleich (Name "n") Gr (Zahl 1))
            [(Zuweisung "f" (Verkn (Name "f") Mal (Name "n"))),
             (Zuweisung "n" (Verkn (Name "n") Minus (Zahl 1)))
            ])
          ]]
  )

z5 :: Zustand
z5 = Akt drStart5

-- f := 1;
-- while (n>1) do
--   f:=(f*n);
--   n:=(n-1);
-- od

-- Fakulttsfunktion, ohne Schleifendurchlauf (0!)

------------------------------------------------------------

-- Anzeigefunktionen fuer Alles:

zeigeOp :: Operator -> String
zeigeOp Plus  = "+"
zeigeOp Minus = "-"
zeigeOp Mal   = "*"
zeigeOp Div   = "`div`"
zeigeOp Mod   = "`mod`"

zeigeVergl :: Vergleicher -> String
zeigeVergl Kl  = "<"
zeigeVergl Gl  = "="
zeigeVergl Gr  = ">"
zeigeVergl Kg  = "<="
zeigeVergl Ug  = "/="
zeigeVergl Gg  = ">="

zeigeBed :: Bedingung -> String
zeigeBed (Wahr True)  = "true"
zeigeBed (Wahr False) = "false"
zeigeBed (Vergleich a1 vergl a2) =
   "("++ (zeigeAusdr a1) ++
         (zeigeVergl vergl) ++
         (zeigeAusdr a2) ++
   ")"

zeigeAusdr :: Ausdruck -> String
zeigeAusdr (Zahl i) = showInt i
zeigeAusdr (Name s) = s
zeigeAusdr (Verkn a1 op a2) =
   "("++ (zeigeAusdr a1) ++
         (zeigeOp op) ++
         (zeigeAusdr a2) ++
   ")"


zeigeAnwf :: [Anweisung] -> String
zeigeAnwf []     = "[]"
zeigeAnwf [a]    = (zeigeAnw a)
zeigeAnwf (a:af) = (zeigeAnw a)++";"++(zeigeAnwf af)

zeigeAnw :: Anweisung -> String
zeigeAnw (Fall b af1 af2) =
   "if "++(zeigeBed b)++
   " then "++(zeigeAnwf af1)++
   " else "++(zeigeAnwf af2)++
   " fi"
zeigeAnw (Schleife b af) =
   "while "++(zeigeBed b)++
   " do "++(zeigeAnwf af)++
   " od"
zeigeAnw (Zuweisung s ausdr) =
   s++":="++(zeigeAusdr ausdr)
zeigeAnw Skip =
   "skip"

zeigeTeile :: [Teil] -> String
zeigeTeile []     = ""
zeigeTeile [t]    = (zeigeTeil t)
zeigeTeile (t:te) = (zeigeTeil t)++"&"++(zeigeTeile te)

zeigeTeil :: Teil -> String
zeigeTeil (Anw a)       = (zeigeAnw a)
zeigeTeil (AnwF af)     = (zeigeAnwf af)
zeigeTeil (Bed b)       = (zeigeBed b)
zeigeTeil (Ausdr a)     = (zeigeAusdr a)
zeigeTeil (Zwert i)     = (showInt i)
zeigeTeil (Bez s)       = s
zeigeTeil (Wwert True)  = "true"
zeigeTeil (Wwert False) = "false"
zeigeTeil (Op op)       = (zeigeOp op)
zeigeTeil (Rop v)       = (zeigeVergl v)
zeigeTeil If            = "if"
zeigeTeil Do            = "do"
zeigeTeil Zuw           = ":="


zeigeWert :: Wert -> String
zeigeWert Undef    = "_"
zeigeWert (Wert i) = showInt i

zeigeVar :: Variable -> String
zeigeVar (Variable s w) = "("++s++","++(zeigeWert w)++")"

zeigeVars :: [Variable] -> String
zeigeVars (v:vars) = (zeigeVar v)++(zeigeVars vars)
zeigeVars []       = ""

   
zeigeKonf :: Konfiguration -> String
zeigeKonf (vars,wk,pk) =
   "<< "++ (zeigeVars vars) ++" , "
      ++ (zeigeTeile (reverse wk)) ++" , "
      ++ (zeigeTeile pk) ++
   " >>"

-- Die Funktion zeigeKonf wird von der Tracerfunktion
-- genutzt. zeigeKonf setzt aus den einzelnen Bestand-
-- teilen des bergebenen Zustands, die aktuelle Umgebung
-- und die beiden Kellerspeicher, einen einzigen String -
-- in spitze Klammern gesetzt - zusammen. Smtliche Zeige-
-- funktionen sind an die vorher eingefhrte Syntax mit den
-- entsprechenden Datentypen des Interpreters angelehnt.

-- Aufruf:   zeigeKonf drStart2
-- Ergebnis: "<< (d,1)(x,17)(y,4)(z,_) ,  , d:=(d+1);x:=(x-y) >>"
-- Aufruf:   zeigeKonf ([(Variable "a" (Wert 1)),(Variable "b" Undef)],[],[Anw (Zuweisung "a" (Zahl 1))])
-- Ergebnis: "<< (a,1)(b,_) ,  , a:=1 >>"

------------------------------------------------------------

-- Zustandsuebergangsfunktion der Abstrakten Maschine
-- (Nummerierung entsprechend der Vorlesungsnotiz 6):

delta :: Zustand -> Zustand
-- 1)
delta (Akt (vars,wk,(Ausdr (Zahl i)):pk)) =
      (Akt (vars,(Zwert i):wk,pk))
-- 2)
delta (Akt (vars,wk,(Ausdr (Name s)):pk)) =
      (Akt (vars,(Zwert i):wk,pk))
        where (Wert i) = (wertSuche s vars)
-- 3)
delta (Akt (vars,wk,(Ausdr(Verkn a1 op a2):pk))) =
      (Akt (vars,wk,(Ausdr a1):(Ausdr a2):(Op op):pk))
-- 4)
delta (Akt (vars,(Zwert n2):(Zwert n1):wk,((Op op):pk))) =
      (Akt (vars,(Zwert ((haskellOp op) n1 n2)):wk,pk))
-- 5)
delta (Akt (vars,wk,(Bed(Wahr tf)):pk)) =
      (Akt (vars,(Wwert tf):wk,pk))
-- 6)
delta (Akt (vars,wk,(Bed(Vergleich a1 rop a2):pk))) =
      (Akt (vars,wk,(Ausdr a1):(Ausdr a2):(Rop rop):pk))
-- 7)
delta (Akt (vars,(Zwert n2):(Zwert n1):wk,((Rop rop):pk))) =
      (Akt (vars,(Wwert ((haskellRop rop) n1 n2)):wk,pk))
-- 8)
delta (Akt (vars,wk,(Anw Skip):pk)) =
      (Akt (vars,wk,pk))
-- 9)
delta (Akt (vars,wk,(Anw (Zuweisung s a)):pk)) =
      (Akt (vars,wk,(Ausdr a):Zuw:(Bez s):pk))
-- 10)
delta (Akt (vars,(Zwert n):wk,Zuw:(Bez s):pk)) =
      (Akt ((wertEintrag s n vars),wk,pk))
-- 11)
delta (Akt (vars,wk,(AnwF (anw:rest)):pk)) =
      (Akt (vars,wk,(Anw anw):(AnwF rest):pk))
-- 11')
delta (Akt (vars,wk,(AnwF []):pk)) =
      (Akt (vars,wk,pk))
-- 12)
delta (Akt (vars,wk,(Anw (Fall b af1 af2)):pk)) =
      (Akt (vars,wk,(Bed b):If:(AnwF af1):(AnwF af2):pk))
-- 13)
delta (Akt (vars,(Wwert True):wk,If:(AnwF af1):_:pk)) =
      (Akt (vars,wk,(AnwF af1):pk))
-- 14)
delta (Akt (vars,(Wwert False):wk,If:_:(AnwF af2):pk)) =
      (Akt (vars,wk,(AnwF af2):pk))
-- 15)
delta (Akt (vars,wk,(Anw (Schleife b af)):pk)) =
      (Akt (vars,wk,(Bed b):Do:(Bed b):(AnwF af):pk))
-- 16)
delta (Akt (vars,(Wwert True):wk,Do:(Bed b):(AnwF af):pk)) =
      (Akt (vars,wk,(AnwF af):(Anw (Schleife b af)):pk))
-- 17)
delta (Akt (vars,(Wwert False):wk,Do:_:_:pk)) =
      (Akt (vars,wk,pk))
-- sonst)
delta (Akt (vars,_,_)) =
            (Ende vars)

-- Die Funktion delta ist das Kernstck der abstrakten
-- Maschine des While-Interpreters. Hier werden die 17
-- Regeln der Zustandsbergangsfunktion ber Pattern 
-- Matching einzeln definiert. Sofern keine Regel mehr 
-- anwendbar ist - also wenn der Programmkeller leer ist
-- - wird der Datentyp Zustand von der Konfiguration auf
-- die aktuelle Umgebung - also die Endspeicherbelegung - 
-- reduziert.

-- Aufruf:   delta (Akt ([(Variable "a" (Wert 1)),
--                        (Variable "b" Undef)],
--                        [],
--                        [Anw (Zuweisung "a" (Zahl 1))
--                       ]))
-- Ergebnis:       (Akt ([(Variable "a" (Wert 1)),
--                        (Variable "b" Undef)],
--                        [],
--                        [(Ausdr (Zahl 1)),Zuw,(Bez "a")
--                       ]))

-- Mit zeigeKonf knnen wir prfen, welcher Zustand von der
-- Zustandsbergangsfunktion ausgegeben wurde:

-- Aufruf:   zeigeKonf ([(Variable "a" (Wert 1)),
--                       (Variable "b" Undef)],
--                       [],
--                       [(Ausdr (Zahl 1)),Zuw,(Bez "a")
--                      ])
-- Ergebnis: "<< (a,1)(b,_) ,  , 1&:=&a >>"

-- Es wurde also Regel 9 angewandt und die Zuweisung wurde
-- aufgelst, um nun den Zuweisungsausdruck separat prfen
-- zu knnen (Regel 1-3 => ob Zahl, Variable oder Ausdruck).

------------------------------------------------------------
-- Uebergaenge kommentiert anzeigen:

regel :: Zustand -> String
regel (Akt (_,_,(Ausdr (Zahl _)):_)) =
      "\n|-- (1 : Zahl lesen)--\n"
regel (Akt (_,_,(Ausdr (Name _)):_)) =
      "\n|-- (2 : Variable lesen, Wert aus Umgebung)--\n"
regel (Akt (_,_,(Ausdr(Verkn _ _ _):_))) =
      "\n|-- (3 : Verknuepfung aufloesen)--\n"
regel (Akt (_,(Zwert _):(Zwert _):_,((Op _):_))) =
      "\n|-- (4 : Arithmetische Operation)--\n"
regel (Akt (_,_,(Bed(Wahr _)):_)) =
      "\n|-- (5 : Wahrheitswert lesen)--\n"
regel (Akt (_,_,(Bed(Vergleich _ _ _):_))) =
      "\n|-- (6 : Bedingung aufloesen)--\n"
regel (Akt (_,(Zwert _):(Zwert _):_,((Rop _):_))) =
      "\n|-- (7 : Vergleichsoperation)--\n"
regel (Akt (_,_,(Anw Skip):_)) =
      "\n|-- (8 : Skipanweisung)--\n"
regel (Akt (_,_,(Anw (Zuweisung _ _)):_)) =
      "\n|-- (9 : Zuweisung aufloesen)--\n"
regel (Akt (_,(Zwert _):_,Zuw:(Bez _):_)) =
      "\n|-- (10 : Zuweisung ausfuehren)--\n"
regel (Akt (_,_,(AnwF (_:_)):_)) =
      "\n|-- (11 : nichtleere Anweisungsfolge aufloesen)--\n"
regel (Akt (_,_,(AnwF []):_)) =
      "\n|-- (11' : leere Anweisungsfolge aufloesen)--\n"
regel (Akt (_,_,(Anw (Fall _ _ _)):_)) =
      "\n|-- (12 : If-Anweisung aufloesen)--\n"
regel (Akt (_,(Wwert True):_,If:(AnwF _):_:_)) =
      "\n|-- (13 : If-Anweisung True-Fall)--\n"
regel (Akt (_,(Wwert False):_,If:_:(AnwF _):_)) =
      "\n|-- (14 : If-Anweisung False-Fall)--\n"
regel (Akt (_,_,(Anw (Schleife _ _)):_)) =
      "\n|-- (15 : Schleife aufloesen)--\n"
regel (Akt (_,(Wwert True):_,Do:(Bed _):(AnwF _):_)) =
      "\n|-- (16 : Schleifenbedingung True)--\n"
regel (Akt (_,(Wwert False):_,Do:_:_:_)) =
      "\n|-- (17 : Schleifenbedingung False)--\n"
regel (Akt (_,_,_)) =
      "\n|-- (Ende der Auswertung)--\n"

-- Die Funktion regel wird vom Tracer verwendet, um die
-- jeweile Regel, welche auf den aktuellen Zustand ausge-
-- fhrt wird, explizit zu nennen und eine kurze Beschreib-
-- ung des Vorgangs zu drucken. Die Fallunterscheidungen
-- unterscheiden sich von der delta Funktion ausschlielich
-- durch smtliche Underscores, da ausschlielich die Form
-- des Zustandes, nicht jedoch die tatschlichen Belegungen
-- von Bedeutung sind. ber die Funktion regel werden die
-- Befehle zum Zeilenumbruch in den Ausgabestring eingefgt.


-- Aufruf:   regel (Akt ([(Variable "a" (Wert 1)),
--                        (Variable "b" Undef)],
--                        [],
--                        [Anw (Zuweisung "a" (Zahl 1))
--                       ]))
-- Ergebnis: "\n|-- (9 : Zuweisung aufloesen)--\n"

-- Traces der Abstrakten Maschine erzeugen ...

trace :: Zustand -> String
trace (Akt k)    = (zeigeKonf k) ++
                   (regel (Akt k)) ++
                   (trace (delta (Akt k)))
trace (Ende umg) = "Endspeicherbelegung: {" ++
                   (zeigeVars umg)++"}"

-- ... und ausdrucken:

zeigeTrace :: Zustand -> IO()
zeigeTrace z = putStr (trace z)

-- Die Funktion trace sorgt in erster Linie fr die
-- rekursive Auswertung des Anfangszustands bis zur End-
-- speicherbelegung. Dabei werden der Anfangszustand, die
-- Zwischenzustnde whrend der Auswertung der Zustands-
-- bergangsfunktion und die Endspeicherbelegung jeweils mit
-- Nennung der jeweiligen angewandten Regel in einem langen
-- String zusammengefasst. Dieser wird anschlieend von der
-- Funktion zeigeTrace ausgedruckt.

------------------------------------------------------

-- zeigeTrace z1:

{-

((d,_)(x,17)(y,4) ,  , d:=0)
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,_)(x,17)(y,4) ,  , d:=0&[])
|-- (9 : Zuweisung aufloesen)--
((d,_)(x,17)(y,4) ,  , 0&:=&d&[])
|-- (1 : Zahl lesen)--
((d,_)(x,17)(y,4) , 0 , :=&d&[])
|-- (10 : Zuweisung ausfuehren)--
((d,0)(x,17)(y,4) ,  , [])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,0)(x,17)(y,4) ,  , )
|-- (Ende der Auswertung)--
Endspeicherbelegung: {(d,0)(x,17)(y,4)}

-}

-- zeigeTrace z2:

{-

((d,1)(x,17)(y,4)(z,_) ,  , d:=(d+1);x:=(x-y))
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,1)(x,17)(y,4)(z,_) ,  , d:=(d+1)&x:=(x-y))
|-- (9 : Zuweisung aufloesen)--
((d,1)(x,17)(y,4)(z,_) ,  , (d+1)&:=&d&x:=(x-y))
|-- (3 : Verknuepfung aufloesen)--
((d,1)(x,17)(y,4)(z,_) ,  , d&1&+&:=&d&x:=(x-y))
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,17)(y,4)(z,_) , 1 , 1&+&:=&d&x:=(x-y))
|-- (1 : Zahl lesen)--
((d,1)(x,17)(y,4)(z,_) , 1&1 , +&:=&d&x:=(x-y))
|-- (4 : Arithmetische Operation)--
((d,1)(x,17)(y,4)(z,_) , 2 , :=&d&x:=(x-y))
|-- (10 : Zuweisung ausfuehren)--
((d,2)(x,17)(y,4)(z,_) ,  , x:=(x-y))
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,2)(x,17)(y,4)(z,_) ,  , x:=(x-y)&[])
|-- (9 : Zuweisung aufloesen)--
((d,2)(x,17)(y,4)(z,_) ,  , (x-y)&:=&x&[])
|-- (3 : Verknuepfung aufloesen)--
((d,2)(x,17)(y,4)(z,_) ,  , x&y&-&:=&x&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,17)(y,4)(z,_) , 17 , y&-&:=&x&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,17)(y,4)(z,_) , 17&4 , -&:=&x&[])
|-- (4 : Arithmetische Operation)--
((d,2)(x,17)(y,4)(z,_) , 13 , :=&x&[])
|-- (10 : Zuweisung ausfuehren)--
((d,2)(x,13)(y,4)(z,_) ,  , [])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,2)(x,13)(y,4)(z,_) ,  , )
|-- (Ende der Auswertung)--
Endspeicherbelegung: {(d,2)(x,13)(y,4)(z,_)}

-}

-- zeigeTrace z3:

{-

((d,_)(x,17)(y,4) ,  , d:=0;while (x>=y) do d:=(d+1);x:=(x-y) od)
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,_)(x,17)(y,4) ,  , d:=0&while (x>=y) do d:=(d+1);x:=(x-y) od)
|-- (9 : Zuweisung aufloesen)--
((d,_)(x,17)(y,4) ,  , 0&:=&d&while (x>=y) do d:=(d+1);x:=(x-y) od)
|-- (1 : Zahl lesen)--
((d,_)(x,17)(y,4) , 0 , :=&d&while (x>=y) do d:=(d+1);x:=(x-y) od)
|-- (10 : Zuweisung ausfuehren)--
((d,0)(x,17)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od)
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,0)(x,17)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (15 : Schleife aufloesen)--
((d,0)(x,17)(y,4) ,  , (x>=y)&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (6 : Bedingung aufloesen)--
((d,0)(x,17)(y,4) ,  , x&y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,0)(x,17)(y,4) , 17 , y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,0)(x,17)(y,4) , 17&4 , >=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (7 : Vergleichsoperation)--
((d,0)(x,17)(y,4) , true , do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (16 : Schleifenbedingung True)--
((d,0)(x,17)(y,4) ,  , d:=(d+1);x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,0)(x,17)(y,4) ,  , d:=(d+1)&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,0)(x,17)(y,4) ,  , (d+1)&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,0)(x,17)(y,4) ,  , d&1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,0)(x,17)(y,4) , 0 , 1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (1 : Zahl lesen)--
((d,0)(x,17)(y,4) , 0&1 , +&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,0)(x,17)(y,4) , 1 , :=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,1)(x,17)(y,4) ,  , x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,1)(x,17)(y,4) ,  , x:=(x-y)&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,1)(x,17)(y,4) ,  , (x-y)&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,1)(x,17)(y,4) ,  , x&y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,17)(y,4) , 17 , y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,17)(y,4) , 17&4 , -&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,1)(x,17)(y,4) , 13 , :=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,1)(x,13)(y,4) ,  , []&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,1)(x,13)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (15 : Schleife aufloesen)--
((d,1)(x,13)(y,4) ,  , (x>=y)&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (6 : Bedingung aufloesen)--
((d,1)(x,13)(y,4) ,  , x&y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,13)(y,4) , 13 , y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,13)(y,4) , 13&4 , >=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (7 : Vergleichsoperation)--
((d,1)(x,13)(y,4) , true , do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (16 : Schleifenbedingung True)--
((d,1)(x,13)(y,4) ,  , d:=(d+1);x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,1)(x,13)(y,4) ,  , d:=(d+1)&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,1)(x,13)(y,4) ,  , (d+1)&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,1)(x,13)(y,4) ,  , d&1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,1)(x,13)(y,4) , 1 , 1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (1 : Zahl lesen)--
((d,1)(x,13)(y,4) , 1&1 , +&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,1)(x,13)(y,4) , 2 , :=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,2)(x,13)(y,4) ,  , x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,2)(x,13)(y,4) ,  , x:=(x-y)&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,2)(x,13)(y,4) ,  , (x-y)&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,2)(x,13)(y,4) ,  , x&y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,13)(y,4) , 13 , y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,13)(y,4) , 13&4 , -&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,2)(x,13)(y,4) , 9 , :=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,2)(x,9)(y,4) ,  , []&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,2)(x,9)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (15 : Schleife aufloesen)--
((d,2)(x,9)(y,4) ,  , (x>=y)&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (6 : Bedingung aufloesen)--
((d,2)(x,9)(y,4) ,  , x&y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,9)(y,4) , 9 , y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,9)(y,4) , 9&4 , >=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (7 : Vergleichsoperation)--
((d,2)(x,9)(y,4) , true , do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (16 : Schleifenbedingung True)--
((d,2)(x,9)(y,4) ,  , d:=(d+1);x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,2)(x,9)(y,4) ,  , d:=(d+1)&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,2)(x,9)(y,4) ,  , (d+1)&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,2)(x,9)(y,4) ,  , d&1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,2)(x,9)(y,4) , 2 , 1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (1 : Zahl lesen)--
((d,2)(x,9)(y,4) , 2&1 , +&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,2)(x,9)(y,4) , 3 , :=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,3)(x,9)(y,4) ,  , x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,3)(x,9)(y,4) ,  , x:=(x-y)&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,3)(x,9)(y,4) ,  , (x-y)&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,3)(x,9)(y,4) ,  , x&y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,3)(x,9)(y,4) , 9 , y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,3)(x,9)(y,4) , 9&4 , -&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,3)(x,9)(y,4) , 5 , :=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,3)(x,5)(y,4) ,  , []&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,3)(x,5)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (15 : Schleife aufloesen)--
((d,3)(x,5)(y,4) ,  , (x>=y)&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (6 : Bedingung aufloesen)--
((d,3)(x,5)(y,4) ,  , x&y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,3)(x,5)(y,4) , 5 , y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,3)(x,5)(y,4) , 5&4 , >=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (7 : Vergleichsoperation)--
((d,3)(x,5)(y,4) , true , do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (16 : Schleifenbedingung True)--
((d,3)(x,5)(y,4) ,  , d:=(d+1);x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,3)(x,5)(y,4) ,  , d:=(d+1)&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,3)(x,5)(y,4) ,  , (d+1)&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,3)(x,5)(y,4) ,  , d&1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,3)(x,5)(y,4) , 3 , 1&+&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (1 : Zahl lesen)--
((d,3)(x,5)(y,4) , 3&1 , +&:=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,3)(x,5)(y,4) , 4 , :=&d&x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,4)(x,5)(y,4) ,  , x:=(x-y)&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11 : nichtleere Anweisungsfolge aufloesen)--
((d,4)(x,5)(y,4) ,  , x:=(x-y)&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (9 : Zuweisung aufloesen)--
((d,4)(x,5)(y,4) ,  , (x-y)&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (3 : Verknuepfung aufloesen)--
((d,4)(x,5)(y,4) ,  , x&y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,4)(x,5)(y,4) , 5 , y&-&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,4)(x,5)(y,4) , 5&4 , -&:=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (4 : Arithmetische Operation)--
((d,4)(x,5)(y,4) , 1 , :=&x&[]&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (10 : Zuweisung ausfuehren)--
((d,4)(x,1)(y,4) ,  , []&while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,4)(x,1)(y,4) ,  , while (x>=y) do d:=(d+1);x:=(x-y) od&[])
|-- (15 : Schleife aufloesen)--
((d,4)(x,1)(y,4) ,  , (x>=y)&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (6 : Bedingung aufloesen)--
((d,4)(x,1)(y,4) ,  , x&y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,4)(x,1)(y,4) , 1 , y&>=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (2 : Variable lesen, Wert aus Umgebung)--
((d,4)(x,1)(y,4) , 1&4 , >=&do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (7 : Vergleichsoperation)--
((d,4)(x,1)(y,4) , false , do&(x>=y)&d:=(d+1);x:=(x-y)&[])
|-- (17 : Schleifenbedingung False)--
((d,4)(x,1)(y,4) ,  , [])
|-- (11' : leere Anweisungsfolge aufloesen)--
((d,4)(x,1)(y,4) ,  , )
|-- (Ende der Auswertung)--
Endspeicherbelegung: {(d,4)(x,1)(y,4)}

-}

